Imports System
Imports System.Runtime.InteropServices.WindowsRuntime
Imports System.Threading
Imports System.Threading.Tasks
Imports System.Windows
Imports System.Windows.Threading
Imports Windows.Networking.Sockets
Imports Windows.Storage.Streams

Namespace WinPhoneFtp.FtpService

    Public NotInheritable Class TcpClientSocket
        Implements IDisposable

        Private IpAddress As String = String.Empty
        Private Port As String = String.Empty
        Private streamSocket As StreamSocket = Nothing
        Private SocketName As String = String.Empty
        Private TcpStreamWriter As DataWriter = Nothing
        Private IsConnectionClosedByRemoteHost As Boolean = False
        Private ReadBufferLength As UInt32 = 0
        Private IsSocketConnected As Boolean = False
        Private logger As Logger = Nothing
        Private UIDispatcher As Dispatcher = Nothing
        Private TcpStreamReader As DataReader = Nothing

        Public Event DataReceived As EventHandler(Of DataReceivedEventArgs)
        Public Event SocketClosed As EventHandler(Of SocketClosedEventArgs)
        Public Event ErrorOccured As EventHandler(Of ErrorOccuredEventArgs)
        Public Event SocketConnected As EventHandler

        Public Sub New(ByVal IpAddress As String, ByVal Port As String, ByVal UIDispatcher As Dispatcher)

            Me.New(IpAddress, Port, 512, "NoName", UIDispatcher)

        End Sub

        Public Sub New(ByVal IpAddress As String, ByVal Port As String, ByVal ReadBufferLength As UInt32, ByVal UIDispatcher As Dispatcher)

            Me.New(IpAddress, Port, ReadBufferLength, "NoName", UIDispatcher)

        End Sub

        Public Sub New(ByVal IpAddress As String, ByVal Port As String, ByVal SocketName As String, ByVal UIDispatcher As Dispatcher)

            Me.New(IpAddress, Port, 512, SocketName, UIDispatcher)

        End Sub

        Public Sub New(ByVal IpAddress As String, ByVal Port As String, ByVal ReadBufferLength As UInt32, ByVal SocketName As String, ByVal UIDispatcher As Dispatcher)

            Me.UIDispatcher = If(UIDispatcher IsNot Nothing, UIDispatcher, Deployment.Current.Dispatcher)
            logger = Logger.GetDefault(UIDispatcher)
            Me.IpAddress = IpAddress
            Me.Port = Port
            Me.SocketName = SocketName
            Me.ReadBufferLength = ReadBufferLength
            'logger.AddLog(String.Format("Creating new TCP Socket {0}. IPAddress: {1}, Port: {2}", SocketName, IpAddress, Port))

        End Sub

        Public Async Function PrepareSocketAsync() As Task

            streamSocket = New StreamSocket

            'logger.AddLog(String.Format("Connecting {0} socket at IPAddress: {1}, Port: {2}", SocketName, IpAddress, Port))
            Try

                Await streamSocket.ConnectAsync(New Windows.Networking.HostName(IpAddress), Port)

            Catch ex As Exception

                'logger.AddLog("Unable to connect to remote connection")
                RaiseErrorOccuredEvent(ex)
                Exit Function

            End Try
            'logger.AddLog("Connected")
            TcpStreamWriter = New DataWriter(streamSocket.OutputStream)
            IsSocketConnected = True
            Try

                Await Task.Factory.StartNew(Sub(stream)

                                                TcpStreamReader = New DataReader(CType(stream, IInputStream))
                                                TcpStreamReader.InputStreamOptions = InputStreamOptions.Partial
                                                Try

                                                    Dim loadOperation As DataReaderLoadOperation = TcpStreamReader.LoadAsync(ReadBufferLength)
                                                    loadOperation.Completed = New Windows.Foundation.AsyncOperationCompletedHandler(Of UInt32)(AddressOf LoadCompleted)

                                                Catch ex As Exception

                                                    RaiseErrorOccuredEvent(ex)

                                                End Try

                                            End Sub, streamSocket.InputStream)

            Catch

                'logger.AddLog("Asynchronous Read Operation Canceled")

            End Try
            RaiseSocketConnectedEvent()

        End Function

        Private Sub LoadCompleted(ByVal asyncInfo As Windows.Foundation.IAsyncOperation(Of UInteger), ByVal asyncStatus As Windows.Foundation.AsyncStatus)

            Select Case asyncStatus

                Case Windows.Foundation.AsyncStatus.Canceled
                    'logger.AddLog("Data load operation canceled")

                Case Windows.Foundation.AsyncStatus.Completed
                    'logger.AddLog("Data load operation completed")
                    If TcpStreamReader.UnconsumedBufferLength.Equals(0) Then

                        If IsSocketConnected Then

                            'logger.AddLog("Connection closed by remote host. Exiting")
                            IsSocketConnected = False
                            IsConnectionClosedByRemoteHost = True
                            CloseSocket()

                        End If

                    Else

                        Dim buffer As IBuffer = TcpStreamReader.DetachBuffer
                        RaiseDataReceivedEvent(buffer.ToArray)
                        Dim loadOperation As DataReaderLoadOperation = TcpStreamReader.LoadAsync(ReadBufferLength)
                        loadOperation.Completed = New Windows.Foundation.AsyncOperationCompletedHandler(Of UInt32)(AddressOf LoadCompleted)

                    End If

                Case Windows.Foundation.AsyncStatus.Error
                    'logger.AddLog("Exception in data load operation")
                    IsSocketConnected = False
                    If asyncInfo.ErrorCode.HResult.Equals(-2147014842) Then

                        IsConnectionClosedByRemoteHost = True

                    Else

                        RaiseErrorOccuredEvent(asyncInfo.ErrorCode)

                    End If
                    CloseSocket()

                Case Windows.Foundation.AsyncStatus.Started
                    'logger.AddLog("Data load operation started")

            End Select

        End Sub

        Public Async Function SendDataAsync(ByVal data() As Byte) As Task

            If TcpStreamWriter IsNot Nothing Then

                TcpStreamWriter.WriteBytes(data)
                Await TcpStreamWriter.StoreAsync

            End If

        End Function

        Public Async Function SendDataAsync(ByVal data As String) As Task

            If TcpStreamWriter IsNot Nothing Then

                TcpStreamWriter.WriteString(data)
                Await TcpStreamWriter.StoreAsync

            End If

        End Function

        <System.Security.SecuritySafeCritical>
        Public Sub CloseSocket()

            IsSocketConnected = False
            'logger.AddLog(String.Format("Closing {0} socket", SocketName))
            If TcpStreamWriter IsNot Nothing Then

                'loadOperation.Close
                TcpStreamWriter.Dispose()
                TcpStreamWriter = Nothing

            End If

            If streamSocket IsNot Nothing Then

                DataReceivedEvent = Nothing
                ErrorOccuredEvent = Nothing
                If Not IsConnectionClosedByRemoteHost Then

                    'loadOperation.Cancel
                    RaiseSocketClosedEvent(SocketCloseReason.ClosedFromLocalHost)

                Else

                    RaiseSocketClosedEvent(SocketCloseReason.ClosedByRemoteHost)

                End If
                SocketClosedEvent = Nothing
                streamSocket.Dispose()
                streamSocket = Nothing

            End If

        End Sub

        Private Sub RaiseDataReceivedEvent(ByVal data() As Byte)

            RaiseEvent DataReceived(Me, New DataReceivedEventArgs(data))

        End Sub

        Private Sub RaiseSocketClosedEvent(ByVal CloseReason As SocketCloseReason)

            RaiseEvent SocketClosed(Me, New SocketClosedEventArgs(CloseReason))

        End Sub

        Private Sub RaiseErrorOccuredEvent(ByVal ExceptionObject As Exception)

            RaiseEvent ErrorOccured(Me, New ErrorOccuredEventArgs(ExceptionObject))

        End Sub

        Private Sub RaiseSocketConnectedEvent()

            RaiseEvent SocketConnected(Me, EventArgs.Empty)

        End Sub

        Public Async Sub Dispose() Implements IDisposable.Dispose

            If TcpStreamWriter IsNot Nothing Then

                Await TcpStreamWriter.FlushAsync
                TcpStreamWriter.Dispose()

            End If

            If streamSocket IsNot Nothing Then

                streamSocket.Dispose()

            End If

        End Sub

    End Class

End Namespace
